(* La syntaxe abstraite des expressions mini-ML *)

type expression =
    Var of string
  | Const of int
  | Op of string
  | Fun of string * expression
  | App of expression * expression
(*  | Paire of expression * expression *)
  | Let of string * expression * expression

(* Substitution d'une variable x par une expression b dans une expression a *)

let rec to_string = function
  | Var s -> s
  | App (t1, t2) -> "(" ^ (to_string t1) ^ " " ^ (to_string t2) ^ ")"
  | Fun (x, t) -> "(fun " ^ x ^ " -> " ^ (to_string t) ^ ")"
  | Const c -> string_of_int c
  | Op  s -> s
  | Let (x,t1,t2) -> "(let " ^ x ^ "=" ^ (to_string t1) ^ " in " ^ (to_string t2) ^ ")"
(*  | Pair (t1,t2) -> "(" ^ (to_string t1) ^ "," ^ (to_string t2) ^ ")" *)


let rec subst a x b =
  match a with
    Var v -> if v = x then b else a
  | Const n -> a
  | Op o -> a
  | Fun(v, a1) -> if v = x then a else Fun(v, subst a1 x b)
  | App(a1, a2) -> App(subst a1 x b, subst a2 x b)
(*  | Paire(a1, a2) -> Paire(subst a1 x b, subst a2 x b) *)
  | Let(v, a1, a2) -> Let(v, subst a1 x b, if v = x then a2 else subst a2 x b)

(* Les fonctions partielles d'valuation sans environement *)

exception Cannot_reduce

(* grand pas sans environement *)
let rec grandPas a =
  match a with
    Const c -> Const c
  | Op o -> Op o
  | Fun(x, a) -> Fun(x, a)
  | App(a1, a2) ->
      begin match grandPas a1 with
        Fun(x, a) -> grandPas (subst a x (grandPas a2))
       | Op "incr" -> let Const n1 = grandPas a2 in Const(n1 + 1)
       | Op _ -> failwith "Operateur inconnu"
       | _ -> raise Cannot_reduce
(*      | Op "+" -> let (Paire(Const n1, Const n2)) = grandPas a2 in Const(n1 + n2)
      | Op "*" -> let (Paire(Const n1, Const n2)) = grandPas a2 in Const(n1 * n2)
      | Op "-" -> let (Paire(Const n1, Const n2)) = grandPas a2 in Const(n1 - n2)
      | Op "fst" -> let (Paire(v1, v2)) = grandPas a2 in v1
      | Op "snd" -> let (Paire(v1, v2)) = grandPas a2 in v2 *)
      end
(*  | Paire(a1, a2) -> Paire(grandPas a1,grandPas a2) *)
  | Let(x, a1, a2) -> grandPas (subst a2 x (grandPas a1))
  | _ -> raise Cannot_reduce


(* Grand pas avec environement *)


(* un environement sera une liste d'association (string,value) *)

type valeur =
  | ValueConst of int
  | ValueOp of string
  | Closure of string * expression * (string * valeur) list

let rec value_to_string = function
  | Closure (x,e,env) ->
    let env'=List.map (function s,v -> s^"=>"^(value_to_string v)^",") env in
     "(fun " ^ x ^ " -> " ^ (to_string e) ^ ")["^(List.fold_left (^) "" env')^"]"
  | ValueConst c -> string_of_int c
  | ValueOp  s -> s


let rec grandPasEnv env a =
  match a with
    Const c -> ValueConst c
  | Op o -> ValueOp o
  | Var x -> begin try List.assoc x env with _ -> raise Cannot_reduce end
  | Fun(x, e) -> Closure(x,e,env)
  | App(a1, a2) ->
      begin match grandPasEnv env a1 with
        Closure(x,e,env') -> grandPasEnv ((x,(grandPasEnv env a2))::env') e
       | ValueOp "incr" -> let ValueConst n1 = grandPasEnv env a2 in ValueConst(n1 + 1)
       | ValueOp _ -> failwith "Operateur inconnu"
       | _ -> raise Cannot_reduce
(*      | Op "+" -> let (Paire(Const n1, Const n2)) = grandPas a2 in Const(n1 + n2)
      | Op "*" -> let (Paire(Const n1, Const n2)) = grandPas a2 in Const(n1 * n2)
      | Op "-" -> let (Paire(Const n1, Const n2)) = grandPas a2 in Const(n1 - n2)
      | Op "fst" -> let (Paire(v1, v2)) = grandPas a2 in v1
      | Op "snd" -> let (Paire(v1, v2)) = grandPas a2 in v2 *)
      end
(*  | Paire(a1, a2) -> Paire(grandPas a1,grandPas a2) *)
  | Let(x, a1, a2) -> grandPasEnv ((x,(grandPasEnv env a1))::env) a2



(* petit pas sans substitution explicite *)

let isvalue = function Const _ -> true | Fun _ -> true | Op _ -> true | _ -> false
let isInt = function Const _ -> true | _ -> false


let rec reduce = function
  | App(Fun(x, a), v) when isvalue v -> subst a x v
  | App(Op "incr", Const n) -> Const (n +1)
  | App(a, b) -> if isvalue a then App(a, reduce b) else App(reduce a, b)
  | Let(x, a, b) -> if isvalue a then subst b x a else Let(x,reduce a,b)
  | _ -> raise Cannot_reduce

let rec petitPas a = try petitPas (reduce a) with Cannot_reduce -> a


